#ifndef _PseudoBooleanLinearizerCallback_hh_
#define _PseudoBooleanLinearizerCallback_hh_

/**
 * defines callback that are used by the parser to tranmit information
 * to the solver
 *
 * These functions must be redefined in a subclass.
 */
template <typename IntegerType>
class PseudoBooleanLinearizerCallback
{
public:
  /**
   * callback called when we get the number of variables and the
   * expected number of constraints
   * @param nbvar: the number of variables
   * @param nbconstr: the number of contraints
   */
  void metaData(int nbvar, int nbconstr)
  {
    cout << "* #variable= " << nbvar 
	 << " #constraint= " << nbconstr  << endl;
  }

  /**
   * callback called before we read the objective function
   */
  void beginObjective()
  {
    cout << "min:";
  }

  /**
   * callback called after we've read the objective function
   */
  void endObjective()
  {
    cout << " ;" << endl;
  }


  /**
   * callback called when we read a term of the objective function
   *
   * @param coeff: the coefficient of the term
   * @param idVar: the numerical identifier of the variable
   */
  void objectiveTerm(IntegerType coeff, int idVar)
  {
    cout << ' ' << showpos << coeff << noshowpos << " x" << idVar;
  }

  /**
   * callback called when we read a term of the objective function
   * which is a product of literals
   *
   * @param coeff: the coefficient of the term
   * @param list: list of literals which appear in the product
   */
  void objectiveProduct(IntegerType coeff, vector<int> list)
  {
    throw runtime_error("unimplemented");
  }
  
  /**
   * callback called before we read a constraint
   */
  void beginConstraint()
  {
  }
  
  /**
   * callback called after we've read a constraint
   */
  void endConstraint()
  {
    cout << " ;" << endl;
  }

  /**
   * callback called when we read a term of a constraint
   *
   * @param coeff: the coefficient of the term
   * @param idVar: the numerical identifier of the variable
   */
  void constraintTerm(IntegerType coeff, int idVar)
  {
    cout << " " << showpos << coeff << noshowpos << " x" << idVar;
  }

  /**
   * callback called when we read a term of a constraint which is a
   * product of literals
   *
   * @param coeff: the coefficient of the term
   * @param list: list of literals which appear in the product
   */
  void constraintProduct(IntegerType coeff, vector<int> list)
  {
    throw runtime_error("unimplemented");
  }

  /**
   * callback called when we read the relational operator of a constraint
   *
   * @param relop: the relational oerator (>= or =)
   */
  void constraintRelOp(string relop)
  {
    cout << " " << relop;
  }

  /**
   * callback called when we read the right term of a constraint (also
   * known as the degree)
   *
   * @param val: the degree of the constraint
   */
  void constraintRightTerm(IntegerType val)
  {
    cout << " " << val;
  }

  /**
   * add the necessary constraints to define newSymbol as equivalent
   * to the product (conjunction) of literals in product.
   */
  void linearizeProduct(int newSymbol, vector<int> product)
  {
    IntegerType r;

    // product => newSymbol (this is a clause) 
    // not x0 or not x1 or ... or not xn or newSymbol
    r=1;
    beginConstraint();
    constraintTerm(1,newSymbol);
    for(int i=0;i<product.size();++i)
      if (product[i]>0)
      {
	constraintTerm(-1,product[i]);
	r-=1;
      }
      else
	constraintTerm(1,-product[i]);
    constraintRelOp(">=");
    constraintRightTerm(r);
    endConstraint();

#ifdef ONLYCLAUSES
    // newSymbol => product translated as
    // not newSymbol of xi (for all i)
    for(int i=0;i<product.size();++i)
    {
      r=0;
      beginConstraint();
      constraintTerm(-1,newSymbol);
      if (product[i]>0)
	constraintTerm(1,product[i]);
      else
      {
	constraintTerm(-1,-product[i]);
	r-=1;
      }
      constraintRelOp(">=");
      constraintRightTerm(r);
      endConstraint();
    }
#else
    // newSymbol => product translated as
    // x0+x1+x3...+xn-n*newSymbol>=0
    r=0;
    beginConstraint();
    constraintTerm(-(int)product.size(),newSymbol);
    for(int i=0;i<product.size();++i)
      if (product[i]>0)
	constraintTerm(1,product[i]);
      else
      {
	constraintTerm(-1,-product[i]);
	r-=1;
      }
    constraintRelOp(">=");
    constraintRightTerm(r);
    endConstraint();
#endif
  }

};


#endif
