#ifndef _PseudoBooleanVerifierCallback_h_
#define _PseudoBooleanVerifierCallback_h_

#include <string>
#include <map>
using namespace std;

class BoolVar
{
private:
  bool assigned;
  bool value;
public:
  BoolVar()
  {
    assigned=false;
    value=false;
  }

  BoolVar &operator =(bool val)
  {
    assigned=true;
    value=val;
  }

  bool isAssigned() const
  {
    return assigned;
  }

  bool getValue() const
  {
    return value;
  }
};

/**
 * callback used by a pseudo boolean parser to check if a given
 * interpretation of literals satisfies every linear pseudo boolean
 * constraint
 */
template <typename IntegerType>
class PseudoBooleanVerifierCallback 
{
private:
  IntegerType leftTerm,rightTerm; // left and right terms of a constraint;
  string relOp; // relational operator of a constraint

  // information on a constraint
  string constraintName; // name of a constraint (if any)

  vector<BoolVar> interpretation; // interpretation of each variable

  IntegerType objectiveValue; // value of the objective function for
			      // this interpretation

  ostringstream constraintOutput; // stores a textual representation
				  // of the current constraint with
				  // information about the variables
				  // value (to be used for debugging
				  // when a constraint is not
				  // satisfied)

  int line;

public:

  void metaData(int nbvar, int nbconstr)
  {
  }

  /**
   * forget the current interpretation
   */
  void clear()
  {
    interpretation.clear();
  }

  /**
   * @return the number of assigned variables in the current
   * interpretation
   */
  int nbAssignedVariables()
  {
    return interpretation.size();
  }


  IntegerType getObjectiveValue()
  {
    return objectiveValue;
  }


  /**
   * registers the truth value of a pseudo boolean variable
   */
  void setValue(int idVar, bool value)
  {
    if (idVar>=interpretation.size())
      interpretation.resize(idVar+1);

    if (interpretation[idVar].isAssigned() && 
	interpretation[idVar].getValue()!=value)
    {
      cerr << "ERROR: variable x" << idVar
	   << " is assigned contradictory values !" << endl;
      exit(1);
    }

    interpretation[idVar]=value;
  }

  /**
   * return the truth value of the variable called variable
   * if positive is false, we return the opposite of the variable
   *
   * when the literal is unassigned, we return false in any case
   * unless mustHaveValue is true in which case we throw an exception
   */
  bool getValue(int literal, bool mustHaveValue=false)
  {
    int idVar=abs(literal);

    if (idVar>=interpretation.size() || 
	!interpretation[idVar].isAssigned())
    {
      // literal is unassigned
      // cout << "warning: unassigned literal " << name << endl;
      if (mustHaveValue)
	throw runtime_error("literal without value in objective function");

      return false;
    }

    if (literal>0)
      return interpretation[idVar].getValue();
    else
      return !interpretation[idVar].getValue();
  }

  /**
   * callback called to indicate the current line number
   */
  virtual void setLineNum(int lineNum)
  {
    line=lineNum;
  }

  /**
   * callback called before we read the objective function
   */
  virtual void beginObjective()
  {
    objectiveValue=0;
  }

  /**
   * callback called before to set if we must minimize or maximize
   * the objective function
   *
   * @param mustBeMinimized: true if the objective function must be
   * minimize, false if it must be maximized
   */
  virtual void objectiveType(bool mustBeMinimized)
  {
    // ignored
  }

  /**
   * callback called when we read a term of the objective function
   *
   * @param coeff: the coefficient of the term
   * @param var: the identifier of the variable
   * @param positiveLiteral: true for a positive literal, false otherwise
   */
  virtual void objectiveTerm(IntegerType coeff, int literal)
  {
    if (getValue(literal))
      objectiveValue+=coeff;
  }

  /**
   * callback called when we read a term of the objective function
   * which is a product of literals
   *
   * @param coeff: the coefficient of the term
   * @param list: list of literals which appear in the product
   */
  void objectiveProduct(IntegerType coeff, vector<int> list)
  {
    int i;
    for(i=0;i<list.size();++i)
    {
      if (!getValue(list[i]))
	break;
    }

    if (i==list.size())
      objectiveValue+=coeff;
  }


  /**
   * callback called when we read a constant in the objective function
   *
   * @param coeff: the constant
   */
  virtual void objectiveConstant(IntegerType coeff)
  {
    objectiveValue+=coeff;
  }

  /**
   * callback called when we need to adjust each coefficient of the
   * previous terms (the ones for which constraintTerm was already
   * called). This can happen for example when we have read integer
   * coefficients so far and the parser discovers a 0.3
   * coefficient. To get all coefficient as integers, each coefficient
   * must be multiplied by 10.
   *
   * @param coeff: the integer by which each previous term must be multiplied
   */
  virtual void multPreviousObjectiveTerms(IntegerType coeff)
  {
    objectiveValue*=coeff;
  }

  /**
   * callback called after we've read the objective function
   */
  virtual void endObjective()
  {
    // ignored
  }


  /**
   * callback called before we read a constraint
   */
  virtual void beginConstraint()
  {
    leftTerm=0;
    rightTerm=0;
    constraintName="";

    constraintOutput.str("");
  }

  /**
   * callback called when we read a term of a constraint
   *
   * @param coeff: the coefficient of the term
   * @param var: the identifier of the variable
   * @param positiveLiteral: true for a positive literal, false otherwise
   */
  virtual void constraintTerm(IntegerType coeff, int literal)
  {
    constraintOutput << showpos << coeff << noshowpos;

    if (literal<0)
      constraintOutput << "*~x" << -literal;
    else
      constraintOutput << "*x" << literal;

    constraintOutput << "[=";

    if (getValue(literal))
    {
      leftTerm+=coeff;
      constraintOutput << "1";
    }
    else
      constraintOutput << "0";

    constraintOutput << "] ";
  }

  /**
   * callback called when we read a term of a constraint which is a
   * product of literals
   *
   * @param coeff: the coefficient of the term
   * @param list: list of literals which appear in the product
   */
  void constraintProduct(IntegerType coeff, vector<int> list)
  {
    bool allTrue=true;

    constraintOutput << showpos << coeff << noshowpos;
   
    for(int i=0;i<list.size();++i)
    {
      if (list[i]<0)
	constraintOutput << "*~x" << -list[i];
      else
	constraintOutput << "*x" << list[i];

      if(getValue(list[i])) 
	constraintOutput << "[=1]";
      else
      {
	allTrue=false;
	constraintOutput << "[=0]";
      }
    }

    if (allTrue)
      leftTerm+=coeff;
  }

  /**
   * callback called when we read a constant on the left side of a constraint
   *
   * @param coeff: the coefficient of the term
   */
  virtual void constraintConstant(IntegerType coeff)
  {
    leftTerm+=coeff;
  }

  /**
   * callback called when we need to adjust each coefficient of the
   * previous terms (the ones for which constraintTerm was already
   * called). This can happen for example when we have read integer
   * coefficients so far and the parser discovers a 0.3
   * coefficient. To get all coefficient as integers, each coefficient
   * must be multiplied by 10.
   *
   * @param coeff: the integer by which each previous term must be multiplied
   */
  virtual void multPreviousConstraintTerms(IntegerType coeff)
  {
    leftTerm*=coeff;
    rightTerm*=coeff;
    constraintOutput << " [multiply each preceding number by "
		     << coeff << "] ";
  }

  /**
   * callback called when we read the relational operator of a constraint
   *
   * @param relop: the relational operator
   */
  virtual void constraintRelOp(string relop)
  {
    relOp=relop;
    constraintOutput << relop << " ";
  }

  /**
   * callback called when we read the right term of a constraint (also
   * known as the degree)
   *
   * @param val: the degree of the constraint
   */
  virtual void constraintRightTerm(IntegerType degree)
  {
    rightTerm=degree;
    constraintOutput << rightTerm << " ";
  }

  /**
   * callback called after we've read a constraint
   */
  virtual void endConstraint()
  {
    bool result=false;

    if (relOp=="=")
      result=leftTerm==rightTerm;
    else
      if (relOp=="<=")
	result=leftTerm<=rightTerm;
      else
	if (relOp=="<")
	  result=leftTerm<rightTerm;
	else
	  if (relOp==">=")
	    result=leftTerm>=rightTerm;
	  else
	    if (relOp==">")
	      result=leftTerm>rightTerm;
	    else
	    {
	      cerr << "INTERNAL ERROR: unexpected relational operator: " 
		   << relOp << endl;
	      exit(127);
	    }

    if (!result)
    {
      cout << "FAILED" << endl;
      cout << "ERROR: unsatisfied constraint on line " << line;
      if (constraintName!="")
	cout << " (contraint name=" << constraintName << ")";
      cout << endl;

      cout << "\t" << leftTerm << relOp << rightTerm << endl;
      cout << endl;
      cout << "Falsified constraint is:" << endl;
      cout << constraintOutput.str() << endl;
      exit(6);
    }
#ifdef debug
    else
    {
      cout << "OK\t" << leftTerm << relOp << rightTerm 
	   << " (contraint name=" << constraintName << ")" << endl;
    }
#endif
  }

  /**
   * callback called at the end of parsing
   */
  virtual void endParsing()
  {
  }

  void linearizeProduct(int newSymbol, vector<int> product)
  {
    throw runtime_error("unimplemented");
  }
};

// Local Variables:
// mode: C++
// End:

#endif

